#! /usr/bin/env python
# -*- coding: utf-8 -*-
""" Furtive - File Integrity Verification System """


import os
import logging

from furtive.hasher import HashDirectory
from furtive.manifest import Manifest


class Furtive(object):
    """ Furtive is an application which stores file state and allows users to
        verify the state in the future. Example use cases include file archives
        and file transport.

        :param base_dir: Base directory to use for the manifest. Can be a full
                         or relative path.
        :type base_dir: str
        :param manifest_path: Path to the manifest file. Can be a full or
                              relative path.
        :type manifest_path: str

    """

    def __init__(self, base_dir, manifest_path):
        self.base_dir = base_dir
        self.manifest_path = manifest_path
        self.manifest = Manifest(self.base_dir, self.manifest_path)

        if os.path.exists(manifest_path):
            self.manifest.load()

    def create(self):
        """ Create and save a new manifest.

            The contents of the new Manfiest() will be saved to
            `manifest_path`.

            :return: None
        """

        self.manifest.create()
        self.manifest.save()

    def compare(self):
        """ Compare the hashes in the database with the current hashes of files
            on the file system.

            :return: Dictionary of added, deleted, and changed files.
            :return type: dict
        """

        if self.manifest.is_empty():
            raise RuntimeError('Manifest %s not loaded.' % self.manifest_path)

        logging.info('Generating temporary updated manifest.')
        current_manifest = Manifest(self.base_dir, '/dev/null')
        current_manifest.create()

        added_files = list(set(current_manifest.manifest) -
                           set(self.manifest.manifest))
        removed_files = list(set(self.manifest.manifest) -
                             set(current_manifest.manifest))
        common_files = list(set(self.manifest.manifest) &
                            set(current_manifest.manifest))
        changed_files = [common_file for common_file in common_files
                         if current_manifest[common_file] !=
                         self.manifest[common_file]]

        return {'removed': removed_files,
                'added': added_files,
                'changed': changed_files}



#! /usr/bin/env python
# -*- coding: utf-8 -*-
""" Manifest of files and their hashes """

import logging

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import SafeLoader as Loader, SafeDumper as Dumper

from furtive.hasher import HashDirectory


class Manifest(object):
    """ Manifest of files and the associated hashes.

        :param directory: directory which will serve as the root for the
                          manifest. All files under the directory will be
                          hashed and added to or compared with the manifest.
        :param type: str
        :param manifest_file: file location of the manifest file. This is the
                              path which will be used for the ``create()`` and
                              ``compare()`` methods. If the file exists, the
                              ``create()`` method will overwrite it.
    """

    def __init__(self, directory, manifest_file):
        self.directory = directory
        self.manifest_file = manifest_file
        self.manifest = None

    def __getitem__(self, hashed_file):
        return self.manifest[hashed_file]

    def create(self):
        """ Creates a new manifest from the directory by calling
            furtive.hasher.HashDirectory() and placing the return dictionary
            in to `Manifest.manifest`.
        """

        self.manifest = HashDirectory(self.directory).hash_files()

    def load(self):
        """ Load a manifest from the manifest file.

            This method will open the manfiest YAML file and load it in to
            the `manifest` object variable.
        """

        logging.debug('Opening %s', self.manifest_file)
        with open(self.manifest_file, 'r') as manifest_file:
            self.manifest = load(manifest_file.read(), Loader=Loader)

    def save(self):
        """ Save the manifest to the manifest file.

            Open a YAML file and dump the contents of the manifest to it.
        """

        logging.info('Saving manifest to %s', self.manifest_file)
        logging.debug('Opening %s', self.manifest_file)
        with open(self.manifest_file, 'w') as manifest_file:
            manifest_file.write(dump(self.manifest,
                                     default_flow_style=False,
                                     Dumper=Dumper))

        logging.debug('Manifest saved')

    def is_empty(self):
        """ Determines if the manifest within memory is empty.

            This simply checks to see if the manifest is None.

            :return: True if manifest is empty, False otherwise.
            :return type: bool
        """

        return self.manifest is None



#!/usr/bin/env python

from distutils.core import setup

__VERSION__ = '0.2.4'

long_description = "See https://furtive.readthedocs.org"

setup(name='Furtive',
      version=__VERSION__,
      description='File Integrity Verification System',
      author='Derrick Bryant',
      author_email='dbryant4@gmail.com',
      long_description=long_description,
      license='MIT',
      packages=['furtive'],
      scripts=['scripts/furtive'],
      url='https://furtive.readthedocs.org',
      install_requires=[
        'PyYAML==3.11',
        'argparse==1.4.0'
      ]
     )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" Manages the hashing of files """

import os
import hashlib
import logging
import multiprocessing


def hash_task(file_path, hash_algorithm='md5'):
    """ Responsible for hashing a file.

        This function reads in the file ``file_path`` in small chuncks the size
        of the hash algorithm's block size in order to avoid running out of
        memory. This means that this function should be able to read any file
        irregardless of the size.

        :param file_path: path of file to hash
        :type file_path: str

        :param hash_algorithm: the hashing algorithm to use. All options
                               available in `hashlib.algorithms` should work.
                               See:
                               https://docs.python.org/2/library/hashlib.html
        :type hash_algorithm: str

        :return: hash of file
        :return type: dict
    """

    with open(file_path, 'r') as file_to_hash:
        logging.debug('Starting Hash of %s', file_path)
        hash_object = hashlib.new(hash_algorithm)
        while True:
            chunk = file_to_hash.read(hash_object.block_size)
            if not chunk:
                break
            hash_object.update(chunk)
        file_hash = hash_object.hexdigest()
        logging.debug('Hash for %s: %s', file_path, file_hash)
    return {file_path: file_hash}


class HashDirectory(object):
    """ Object to manage hashing files in a directory.

        This object is responsible for walking the directory tree and
        adding each file to a list. Once the directory walk has compelted, each
        file path is passed to hash_task(). After each file has been
        hashed, this object will then create a Python dictionary of files with
        their associated hash.

        :param directory: Path to directory containing files
        :type directory: str

        :return: Dictionary of file:hash
        :return type: dict
    """

    def __init__(self, directory):
        self.directory = directory
        self.hashes = {}

    def hash_files(self):
        """ Orchestrates the discovery and hashing of files.



            Note: This method only supports the md5 hashing algorithm
        """

        files_to_hash = []
        num_processes = multiprocessing.cpu_count()

        logging.info('Discovering files in %s and adding to processing queue',
                     self.directory)
        for root, _, files in os.walk(self.directory):
            for found_file in files:
                full_path = os.path.join(root, found_file)
                relative_path = os.path.relpath(full_path, self.directory)
                logging.debug('Found %s', relative_path)
                files_to_hash.append(relative_path)

        logging.debug('Switching current working directory to %s',
                      self.directory)
        old_cwd = os.getcwd()
        os.chdir(self.directory)
        logging.debug('Starting %s hash worker processes', num_processes)
        pool = multiprocessing.Pool(processes=num_processes)
        logging.info('Hashing %s files', len(files_to_hash))
        results = pool.map(hash_task, files_to_hash)
        os.chdir(old_cwd)

        self.hashes = {}
        for item in results:
            self.hashes[item.keys()[0]] = item.values()[0]

        logging.debug('Ensuring processes have stopped')
        pool.close()
        pool.join()

        return self.hashes
